from .Range import Range
import time
import datetime
import pandas as pd

def load_stream_2_range(stream_data: list, normal_label: int, anomaly_label: int, is_range_name: bool) -> list:
    return_list = []
    start_id = -1
    end_id = -1
    id = 0
    range_id = 1

    prev_val = -2 #Set prev_val as a value different to normal and anomalous labels

    for val in stream_data:
        if val == anomaly_label and (prev_val == normal_label or prev_val == -2): #Enter the anomaly range
            start_id = id
        elif val == normal_label and prev_val == anomaly_label: #Go out the anomaly range
            name_buf = ''
            if is_range_name:
                name_buf = str(range_id)
            end_id = id - 1
            return_list.append(Range(start_id, end_id, name_buf))
            range_id += 1
            #start_id = 0

        id += 1
        prev_val = val
    if start_id > end_id: #start_id != 0 and start_id != -1: #if an anomaly continues till the last point
        return_list.append(Range.Range(start_id, id - 1, str(range_id)))

    return return_list


def load_stream_file(filename: str, normal_label: int, anomaly_label: int, is_range_name: bool) -> list:
    return_list = []
    start_id = -1
    end_id = -1
    id = 0
    range_id = 1
    #is_first = True

    prev_val = -2 #Set prev_val as a value different to normal and anomalous labels

    f = open(filename, 'r', encoding='utf-8', newline='')

    for line in f.readlines():
        val = int(line.strip().split()[0])

        '''
        #skip the first line
        if is_first:
            if val == anomaly_label:
                start_id = id
            prev_val = val
            is_first = False
            continue
        '''

        if val == anomaly_label and (prev_val == normal_label or prev_val == -2): #Enter the anomaly range
            start_id = id
        elif val == normal_label and prev_val == anomaly_label: #Go out the anomaly range
            name_buf = ''
            if is_range_name:
                name_buf = str(range_id)
            end_id = id - 1
            return_list.append(Range.Range(start_id, end_id, name_buf))
            range_id += 1
            #start_id = 0

        id += 1
        prev_val = val
    f.close()
    if start_id > end_id: #start_id != 0 and start_id != -1: #if an anomaly continues till the last point
        return_list.append(Range.Range(start_id, id - 1, str(range_id)))

    return return_list

def load_range_file(filename: str, time_format: str) -> list:
    return_list = []
    #is_first = True

    f = open(filename, 'r', encoding='utf-8', newline='')
    for line in f.readlines():
        # skip the first line
        #if is_first:
            #is_first = False
            #continue

        items = line.strip().split(',')
        if time_format == 'index':
            first_idx = int(items[0])
            last_idx = int(items[1])
        else:
            first_idx = string_to_unixtime(items[0], time_format)
            last_idx = string_to_unixtime(items[1], time_format)

        name_buf = ''
        if len(items) > 2:
            name_buf = str(items[2])

        return_list.append(Range.Range(first_idx, last_idx, name_buf))
    f.close()

    for idx in range(1, len(return_list)):
        if return_list[idx].get_time()[0] <= return_list[idx-1].get_time()[1]:
            print("Error: ranges ({},{}) and ({},{}) are overlapped in {}".format(return_list[idx-1].get_time()[0],
                                                                                  return_list[idx-1].get_time()[1],
                                                                                  return_list[idx].get_time()[0],
                                                                                  return_list[idx].get_time()[1], filename))
            exit(0)

    return return_list


def unixtime_to_string(epoch: int, format: str) -> str:
    return datetime.datetime.fromtimestamp(epoch).strftime(format) #'%Y-%m-%d %I:%M:%S %p'


def string_to_unixtime(timestamp: str, format: str) -> int:
    return int(time.mktime(datetime.datetime.strptime(timestamp, format).timetuple()))


def save_range_list(filename: str, range_list: list) -> None:
    f = open(filename, encoding='utf-8', mode='w')
    for single_range in range_list:
        first, last = single_range.get_time()
        f.writelines(str(first)+','+str(last)+','+single_range.get_name()+'\n')
    f.close()

# Assume that the first line of input files including the information of file format and its corresponding information
# This function handles three types of file format
def load_file(filename: str, filetype: str) -> list:
    assert(filetype == 'range' or filetype == 'stream')

    if filetype == 'stream':
        return load_stream_file(filename, 1, -1, True)
    elif filetype == 'range':
        return load_range_file(filename, 'index')


def make_attack_file(input_files: list, sep: str, label_featname: str, input_normal_label: int, input_anomalous_label: int,
                     output_stream_file: str, output_range_file: str, output_normal_label: int, output_anomalous_label: int) -> None:
    label = []
    for an_input_file in input_files:
        temp_file = pd.read_csv(an_input_file, sep=sep)
        label += temp_file[label_featname].values.tolist()

    with open(output_stream_file, 'w') as f:
        for a_label in label:
            if a_label == input_normal_label:
                f.write('{}\n'.format(output_normal_label))
            elif a_label == input_anomalous_label:
                f.write('{}\n'.format(output_anomalous_label))
            else:
                print("There is an unknown label, " + a_label, flush=True)
                f.close()
                return

    ranges = load_stream_2_range(label, 0, 1, False)
    save_range_list(output_range_file, ranges)

def save_range_2_stream(filename: str, range_list: list, last_idx: int, normal_label: int, anomalous_label: int) -> None:
    f = open(filename, encoding='utf-8', mode='w')
    range_id = 0
    for idx in range(last_idx):
        if idx < range_list[range_id].get_time()[0]:
            f.writelines('{}\n'.format(normal_label))
        elif range_list[range_id].get_time()[0] <= idx <= range_list[range_id].get_time()[1]:
            f.writelines('{}\n'.format(anomalous_label))
        else:
            f.writelines('{}\n'.format(normal_label))
            if range_id < len(range_list) - 1:
                range_id += 1
    f.close()


import numpy as np

from ..base.time_series_metrics import BaseTimeSeriesMetrics
from .precision import TimeSeriesPrecision
from .recall import TimeSeriesRecall


class TimeSeriesFScore(BaseTimeSeriesMetrics):
    """ This class calculates f-score for time series"""

    def __init__(self, beta=1.0, p_alpha=0.0, r_alpha=0.0, cardinality="one", p_bias="flat", r_bias="flat"):
        """Constructor

        Args:
            beta (float, optional): determines the weight of recall in the combined score.. Defaults to 1.0.
            p_alpha (float, optional): alpha of precision, 0<=alpha_p<=1. Defaults to 0.0.
            r_alpha (float, optional): alpha of recall, 0<=alpha<=1. Defaults to 0.0.
            cardinality (str, optional): ["one", "reciprocal", "udf_gamma"]. Defaults to "one".
            p_bias (str, optional): bias of precision, ["flat", "front", "middle", "back"]. Defaults to "flat".
            r_bias (str, optional): bias of recall, ["flat", "front", "middle", "back"]. Defaults to "flat".
        """

        assert beta >= 0

        self.beta = beta
        self.p_alpha = p_alpha
        self.r_alpha = r_alpha
        self.cardinality = cardinality
        self.p_bias = p_bias
        self.r_bias = r_bias

    def score(self, real: np.ndarray, pred: np.ndarray) -> float:
        """Computing fbeta score

        Args:
            real (np.ndarray):
                One-dimensional array of correct answers with values of 1 or 0.
            pred (np.ndarray):
                One-dimensional array of predicted answers with values of 1 or 0.

        Returns:
            float: fbeta
        """

        assert isinstance(real, np.ndarray) or isinstance(real, list)
        assert isinstance(pred, np.ndarray) or isinstance(pred, list)

        if not isinstance(real, np.ndarray):
            real = np.array(real)
        if not isinstance(pred, np.ndarray):
            pred = np.array(pred)

        precision = TimeSeriesPrecision(self.p_alpha, self.cardinality, self.p_bias).score(real, pred)
        recall = TimeSeriesRecall(self.r_alpha, self.cardinality, self.r_bias).score(real, pred)

        if precision + recall != 0:
            f_beta = (1 + self.beta ** 2) * precision * recall / (self.beta ** 2 * precision + recall)
        else:
            f_beta = 0

        return f_beta



import numpy as np
from dataclasses import dataclass
from sklearn.metrics import roc_auc_score, average_precision_score

from .nabscore import Sweeper
from .affiliation.metrics import pr_from_events as affiliation_pr
from .prts.basic_metrics_ts import ts_recall, ts_precision, ts_fscore
from .time_tolerant import time_tolerant_recall_,time_tolerant_precision_
from .latency_sparsity_aware import calc_twseq
from .eTaPR_pkg import etapr, tapr
from .eTaPR_pkg.DataManage import File_IO, Range as rng
from .vus_utils import generate_curve


# NOTE:
# Binary anomaly time series (either labels or predictions) are represented in 3 different ways.
# This is done to suit the different metrics.
#
# Example:
# A time series of length 10 (t=0 to t=9)  with anomalies at times t=2, t=6 and t=7 is represented like this:
# Segmentwise: [[2,2], [6,7]]
# Pointwise: [2,6,7]
# Full_series: [0,0,1,0,0,0,1,1,0,0]
#
# The class Binary_anomalies is used to access these various formats.


def pointwise_to_segmentwise(pointwise):
    """Reformat anomaly time series from pointwise to segmentwise"""
    segmentwise = []

    prev = -10
    for point in pointwise:
        if point > prev + 1:
            segmentwise.append([point, point])
        else:
            segmentwise[-1][-1] += 1
        prev = point
    return np.array(segmentwise)


def segmentwise_to_pointwise(segmentwise):
    """Reformat anomaly time series from segmentwise to pointwise"""
    pointwise = []

    for start, end in segmentwise:
        for point in range(start, end + 1):
            pointwise.append(point)

    return np.array(pointwise)


def segmentwise_to_full_series(segmentwise, length):
    """Reformat anomaly time series from segmentwise to full_series"""
    pw = segmentwise_to_pointwise(segmentwise)

    return  pointwise_to_full_series(pw, length)

def pointwise_to_full_series(pointwise, length):
    """Reformat anomaly time series from pointwise to full_series"""
    anomalies_full_series = np.zeros(length)
    if len(pointwise) > 0:
        assert pointwise[-1] < length
        anomalies_full_series[pointwise] = 1
    return np.array(anomalies_full_series)

def full_series_to_pointwise(full_series):
    """Reformat anomaly time series from full_series to pointwise"""
    anomalies_pointwise = []
    for i in range(len(full_series)):
        if full_series[i]==1:
            anomalies_pointwise.append(i)
    return np.array(anomalies_pointwise)

class Binary_anomalies:
    def __init__(self, length, anomalies):
        self._length = length
        self._set_anomalies(anomalies)

    def _set_anomalies(self, anomalies):
        anomalies = np.array(anomalies)
        if self._is_pointwise(anomalies):
            anomalies_ptwise = anomalies

            anomalies_segmentwise = pointwise_to_segmentwise(anomalies)
            
            
            anomalies_full_series = pointwise_to_full_series(anomalies_ptwise, self._length)

            
        elif self._is_full_series(anomalies):
            anomalies_ptwise = full_series_to_pointwise(anomalies)
            anomalies_segmentwise = pointwise_to_segmentwise(anomalies_ptwise)

            anomalies_full_series = anomalies
            tam_pointwiuse = len(np.unique(anomalies_ptwise))
            tam_fullseries = np.sum(anomalies_full_series)
            if tam_pointwiuse != tam_fullseries:
                print(anomalies)
                print(f'2- ERROR EN EL TAMAÑO DE LOS PUNTOS: {tam_pointwiuse} != {tam_fullseries}')
            #raise NotImplementedError
        elif self._is_segmentwise(anomalies):
            anomalies_segmentwise = anomalies
            anomalies_ptwise = segmentwise_to_pointwise(anomalies)
            anomalies_full_series = pointwise_to_full_series(anomalies_ptwise, self._length)
            tam_pointwiuse = len(np.unique(anomalies_ptwise))
            tam_fullseries = np.sum(anomalies_full_series)
            if tam_pointwiuse != tam_fullseries:
                print(f'3- ERROR EN EL TAMAÑO DE LOS PUNTOS: {tam_pointwiuse} != {tam_fullseries}')
        else:
            raise ValueError(f"Illegal shape of anomalies:\n{anomalies}")
        

        if len(anomalies_ptwise) > 0:
            assert all(anomalies_ptwise == np.sort(anomalies_ptwise))
            assert anomalies_ptwise[0] >= 0
            assert len(anomalies_ptwise) == len(np.unique(anomalies_ptwise))
            assert len(anomalies_ptwise) == sum(anomalies_full_series)

            assert all(anomalies_segmentwise[:, 0] == np.sort(anomalies_segmentwise[:, 0]))
            assert all(anomalies_segmentwise[:, 1] >= anomalies_segmentwise[:, 0])

        self.anomalies_segmentwise = anomalies_segmentwise
        self.anomalies_ptwise = anomalies_ptwise
        self.anomalies_full_series = anomalies_full_series

    def _is_pointwise(self, anomalies):
        return len(anomalies.shape) == 1 and (len(anomalies) < self._length or (len(anomalies) == self._length and np.sum(anomalies) > self._length))

    def _is_full_series(self, anomalies):
        return len(anomalies.shape) == 1 and len(anomalies) == self._length

    def _is_segmentwise(self, anomalies):
        return len(anomalies.shape) == 2

    def get_length(self):
        return self._length


class Binary_detection:
    """This class represents a binary detection as a set of two time series:
    gt: the binary labels
    prediction: the binary predictions for corresponding to the labels"""

    def __init__(self, length, gt_anomalies, predicted_anomalies):
        self._length = length
        self._gt = Binary_anomalies(length, gt_anomalies)
        self._prediction = Binary_anomalies(length, predicted_anomalies)

    def get_length(self):
        return self._length

    def get_gt_anomalies_ptwise(self):
        return self._gt.anomalies_ptwise

    def get_gt_anomalies_segmentwise(self):
        return self._gt.anomalies_segmentwise

    def get_predicted_anomalies_ptwise(self):
        return self._prediction.anomalies_ptwise

    def get_predicted_anomalies_segmentwise(self):
        return self._prediction.anomalies_segmentwise

    def get_predicted_anomalies_full_series(self):
        return self._prediction.anomalies_full_series

    def get_gt_anomalies_full_series(self):
        return self._gt.anomalies_full_series


class Nonbinary_detection:
    """This class represents a nonbinary detection as a set of two time series:
    gt: the binary labels
    anomaly score: the time series defining the degree of anomaly at each time point"""

    def __init__(self, gt_anomalies, anomaly_score):
        self._length = len(anomaly_score)
        self._gt = Binary_anomalies(self._length, gt_anomalies)
        self._anomaly_score = anomaly_score

    def get_gt_anomalies_ptwise(self):
        return self._gt.anomalies_ptwise

    def get_gt_anomalies_segmentwise(self):
        return self._gt.anomalies_segmentwise

    def get_gt_anomalies_full_series(self):
        return self._gt.anomalies_full_series

    def get_anomaly_score(self):
        return self._anomaly_score


def f1_from_pr(p, r, beta=1):
    if r == 0 and p == 0:
        return 0
    return ((1 + beta**2) * r * p) / (beta**2 * p + r)


def f1_score(*args, tp, fp, fn, beta=1):
    r = recall(tp=tp, fn=fn)
    p = precision(tp=tp, fp=fp)
    return f1_from_pr(p, r, beta=beta)


def recall(*args, tp, fn):
    return 0 if tp + fn == 0 else tp / (tp + fn)


def precision(*args, tp, fp):
    return 0 if tp + fp == 0 else tp / (tp + fp)


class Pointwise_metrics(Binary_detection):
    def __init__(self, *args):
        super().__init__(*args)
        self.name = "\\pwf[1]"
        self.set_confusion()

    def set_confusion(self):
        gt = self.get_gt_anomalies_full_series()
        pred = self.get_predicted_anomalies_full_series()

        self.tp = np.sum(pred * gt)
        self.fp = np.sum(pred * (1 - gt))
        self.fn = np.sum((1 - pred) * gt)

    def get_score(self):
        return f1_score(tp=self.tp, fn=self.fn, fp=self.fp)


class DelayThresholdedPointAdjust(Pointwise_metrics):
    def __init__(self, *args, k=2):
        super().__init__(*args)
        self.name = f"\\dtpaf[1]{{{k}}}"
        self.k = k
        self.adjust()
        self.set_confusion()

    def adjust(self):
        adjusted_prediction = np.zeros(self._length)
        predicted_anomalies = self.get_predicted_anomalies_ptwise()
        adjusted_prediction[predicted_anomalies] = 1

        for start, end in self.get_gt_anomalies_segmentwise():
            anomaly_adjusted = False
            for i in range(start, min(start + self.k + 1, end + 1)):
                if adjusted_prediction[i] == 1:
                    adjusted_prediction[start:end + 1] = 1
                    anomaly_adjusted = True
                    break
            if not anomaly_adjusted:
                adjusted_prediction[start:end + 1] = 0
        self._prediction._set_anomalies(np.where(adjusted_prediction == 1)[0])


class PointAdjust(DelayThresholdedPointAdjust):
    def __init__(self, *args):
        super().__init__(*args, k=args[0])  # set k to length of time series to avoid threshold making a difference
        self.name = "\\paf[1]"


class PointAdjustKPercent(Pointwise_metrics):
    def __init__(self, *args, k=0.2):
        super().__init__(*args)
        self.name = f"\\pakf[1]{{{int(k*100)}}}"
        self.k = k
        self.adjust()
        self.set_confusion()

    def adjust(self):
        adjusted_prediction = self.get_predicted_anomalies_ptwise().tolist()
        for start, end in self.get_gt_anomalies_segmentwise():
            correct_points = 0
            for i in range(start, end + 1):
                if i in adjusted_prediction:
                    correct_points += 1
                    if correct_points / (end + 1 - start) >= self.k:
                        for j in range(start, end + 1):
                            adjusted_prediction.append(j)
                        break

        self._prediction._set_anomalies(np.sort(np.unique(adjusted_prediction)))


class LatencySparsityAware(Binary_detection):
    def __init__(self, *args, tw=2):
        self.name = f"\\lsf[1]{{{tw}}}"
        super().__init__(*args)
        self.tw = tw
        self.get_score()

    def get_score(self):
        f1, p, r, FPR, self.tp, self.tn, self.fp, self.fn = calc_twseq(
            self.get_predicted_anomalies_full_series(),
            self.get_gt_anomalies_full_series(),
            normal=0,
            threshold=0.5,
            tw=self.tw,
        )
        return f1


class Segmentwise_metrics(Pointwise_metrics):
    def __init__(self, *args):
        super().__init__(*args)
        self.name = "\\segf[1]"
        self.set_confusion()

    def set_confusion(self):
        tp = 0
        fn = 0
        p_fs = self.get_predicted_anomalies_full_series()
        gt_fs = self.get_gt_anomalies_full_series()
        for gt_anomaly in self.get_gt_anomalies_segmentwise():
            found = False
            
            for i_index in range(gt_anomaly[0],gt_anomaly[1]+1):
                if p_fs[i_index] == 1:
                    tp += 1
                    found = True
                    break
            if found == False:
                fn += 1
        fp = 0
        for predicted_anomaly in self.get_predicted_anomalies_segmentwise():
            found = False
            for i_index in range(predicted_anomaly[0],predicted_anomaly[1]+1):
                if gt_fs[i_index] == 1:
                    found = True
                    break
            if found == False:
                fp += 1
        self.fp = fp
        self.fn = fn
        self.tp = tp

    def _overlap(self, anomaly1, anomaly2):
        return not (anomaly1[1] < anomaly2[0] or anomaly2[1] < anomaly1[0])


class Redefined_PR_metric(Binary_detection):
    def __init__(self, *args):
        super().__init__(*args)

    def get_score(self):
        self.r = self.recall()
        self.p = self.precision()
        return f1_from_pr(self.p, self.r)

    def recall(self):
        raise NotImplementedError

    def precision(self):
        raise NotImplementedError


class Composite_f(Redefined_PR_metric):
    def __init__(self, *args):
        self.name = "\\cf[1]"
        super().__init__(*args)

        self.pointwise_metrics = Pointwise_metrics(*args)
        self.segmentwise_metrics = Segmentwise_metrics(*args)

    def recall(self):
        return recall(tp=self.segmentwise_metrics.tp, fn=self.segmentwise_metrics.fn)

    def precision(self):
        return precision(tp=self.pointwise_metrics.tp, fp=self.pointwise_metrics.fp)


class Affiliation(Redefined_PR_metric):
    def __init__(self, *args):
        self.name = "\\af[1]"
        super().__init__(*args)

    def get_score(self, beta=1):
        pr_output = affiliation_pr(
            self._reformat_segments(self.get_predicted_anomalies_segmentwise()),
            self._reformat_segments(self.get_gt_anomalies_segmentwise()),
            (0, self.get_length()),
        )
        self.r = pr_output["recall"]
        self.p = pr_output["precision"]
        return f1_from_pr(self.p, self.r, beta=beta)

    def _reformat_segments(self, segments):
        segments = self._include_end_of_segments(segments)
        segments = self._tuplify_segments(segments)
        return segments

    def _include_end_of_segments(self, segments):
        return [[start, end + 1] for start, end in segments]

    def _tuplify_segments(self, segments):
        return [tuple(segment) for segment in segments]


class Range_PR(Redefined_PR_metric):
    def __init__(self, *args, cardinality= "alpha", alpha=0.2, bias="flat"):
        super().__init__(*args)
        self.cardinality = cardinality
        self.alpha = alpha
        self.bias = bias
        self.set_name()

    def set_name(self):
        self.name = f"\\rf[1]{{{self.bias}}}{{{self.alpha}}}"

    def set_kwargs(self):
        real = np.zeros(self.get_length())
        real[self.get_gt_anomalies_ptwise()] = 1
        pred = np.zeros(self.get_length())
        pred[self.get_predicted_anomalies_ptwise()] = 1

        self.kwargs = {"real": real, "pred": pred, "alpha": self.alpha, "cardinality": self.cardinality, "bias": self.bias}

    def recall(self):
        self.set_kwargs()
        return ts_recall(**self.kwargs)

    def precision(self):
        self.set_kwargs()
        return ts_precision(**self.kwargs)
    



class TaF(Redefined_PR_metric):
    def __init__(self, *args, theta=0.5, alpha=0.5, delta=0, past_range=False):
        super().__init__(*args)
        self.alpha = alpha
        self.theta = theta
        self.delta = delta
        self.past_range = past_range
        self.name = f"\\taf[1]{{{self.alpha}}}{{{self.delta}}}{{{self.theta}}}"

        self.prepare_scoring()

    def prepare_scoring(self):
        #self.prepare_data()
        self.TaPR = tapr.TaPR(theta=self.theta, delta=self.delta,past_range=self.past_range)
        aux_gt_anomalies = []
        for start, end in self.get_gt_anomalies_segmentwise():
            aux_gt_anomalies.append(rng.Range(start, end, ""))
        
        aux_predicted_anomalies = []
        for start, end in self.get_predicted_anomalies_segmentwise():
            aux_predicted_anomalies.append(rng.Range(start, end, ""))
        self.TaPR.set_anomalies(aux_gt_anomalies)
        self.TaPR.set_predictions(aux_predicted_anomalies)

    def prepare_data(self):
        # self.write_data_files()
        # self.read_data_files()
        self.gt_anomalies = self.get_gt_anomalies_full_series().tolist()
        self.predicted_anomalies = self.get_predicted_anomalies_full_series().tolist()

    def write_data_files(self):
        self.gt_filename = "temp_gt.txt"
        with open(self.gt_filename, "w") as f:
            for x in self.get_gt_anomalies_full_series():
                f.write(str(1 if x == 0 else -1))
                f.write("\n")
        self.pred_filename = "temp_pred.txt"
        with open(self.pred_filename, "w") as f:
            for x in self.get_predicted_anomalies_full_series():
                f.write(str(1 if x == 0 else -1))
                f.write("\n")

    def read_data_files(self):
        self.gt_anomalies = File_IO.load_file(self.gt_filename, "stream")
        self.predicted_anomalies = File_IO.load_file(self.pred_filename, "stream")

    def recall(self):
        tard_value, detected_list = self.TaPR.TaR_d()
        tarp_value = self.TaPR.TaR_p_value
        return self.alpha * tard_value + (1 - self.alpha) * tarp_value

    def precision(self):
        tapd_value, correct_list = self.TaPR.TaP_d()
        tapp_value = self.TaPR.TaP_p_value
        return self.alpha * tapd_value + (1 - self.alpha) * tapp_value


class eTaF(Redefined_PR_metric):
    def __init__(self, *args, theta_p=0.5, theta_r=0.1):
        super().__init__(*args)
        self.theta_p = theta_p
        self.theta_r = theta_r

        self.name = f"\\etaf[1]{{{self.theta_p}}}{{{self.theta_r}}}"

        self.make_scores()

    def make_scores(self):
        #self.prepare_data()
        aux_gt_anomalies = []
        for start, end in self.get_gt_anomalies_segmentwise():
            aux_gt_anomalies.append(rng.Range(start, end, ""))
        
        aux_predicted_anomalies = []
        for start, end in self.get_predicted_anomalies_segmentwise():
            aux_predicted_anomalies.append(rng.Range(start, end, ""))
        self.gt_anomalies = aux_gt_anomalies
        self.predicted_anomalies = aux_predicted_anomalies
        self.result = etapr.evaluate_w_ranges(
            self.gt_anomalies, self.predicted_anomalies, theta_p=self.theta_p, theta_r=self.theta_r, delta=0
        )

    def prepare_data(self):
        self.write_data_files()
        self.read_data_files()

    def write_data_files(self):
        self.gt_filename = "temp_gt.txt"
        with open(self.gt_filename, "w") as f:
            for x in self.get_gt_anomalies_full_series():
                f.write(str(1 if x == 0 else -1))
                f.write("\n")
        self.pred_filename = "temp_pred.txt"
        with open(self.pred_filename, "w") as f:
            for x in self.get_predicted_anomalies_full_series():
                f.write(str(1 if x == 0 else -1))
                f.write("\n")

    def read_data_files(self):
        self.gt_anomalies = File_IO.load_file(self.gt_filename, "stream")
        self.predicted_anomalies = File_IO.load_file(self.pred_filename, "stream")

    def recall(self):
        return self.result["eTaR"]

    def precision(self):
        return self.result["eTaP"]


class Time_Tolerant(Redefined_PR_metric):
    def __init__(self, *args, d=2):
        super().__init__(*args)
        self.d = d
        self.name = f"\\ttolf[1]{{{d}}}"

    def recall(self):
        return time_tolerant_recall_(**self.get_kwargs())

    def precision(self):
        return time_tolerant_precision_(**self.get_kwargs())

    def get_kwargs(self):
        return {
            "A": np.pad(self.get_predicted_anomalies_full_series(), self.d),
            "E": np.pad(self.get_gt_anomalies_full_series(), self.d),
            "d": self.d,
        }


class Temporal_Distance(Binary_detection):
    def __init__(self, *args, distance=0):
        super().__init__(*args)
        self.distance = distance
        self.name = f"\\tempdist"

    def get_score(self):
        a = np.array(self.get_gt_anomalies_ptwise())
        b = np.array(self.get_predicted_anomalies_ptwise())
        if self.distance == 0:
            return self._dist(a, b) + self._dist(b, a)
        elif self.distance == 1:
            return self._dist(a, b)**2 + self._dist(b, a)**2
        else:
            raise ValueError(f"Distance {self.distance} not supported")

    def _dist(self, a, b):
        dist = 0
        for pt in a:
            if len(b) > 0:
                dist += min(abs(b - pt))
            else:
                dist += self._length
        return dist


class NAB_score(Binary_detection):
    def __init__(self, *args):
        self.name = "\\nab"
        super().__init__(*args)

        self.sweeper = Sweeper(probationPercent=0, costMatrix={"tpWeight": 1, "fpWeight": 0.11, "fnWeight": 1})

    def get_score(self):
        if len(self.get_predicted_anomalies_ptwise()) == 0:
            return 0  # raw_score == null_score yeilds score = 0
        if len(self.get_gt_anomalies_ptwise()) == 0:
            return np.nan  # perfect_score == null_score yields /0
        try:
            null_score, raw_score = self.calculate_scores(self.get_predicted_anomalies_ptwise())
            null_score, perfect_score = self.calculate_scores(prediction=self.get_gt_anomalies_ptwise())
            return (raw_score - null_score) / (perfect_score - null_score) * 100
        except Exception as e:
            #print(f"Error calculating NAB score: {e}")
            return 0

    def calculate_scores(self, prediction):
        anomaly_scores = pointwise_to_full_series(prediction, self.get_length())
        timestamps = np.arange(self.get_length())
        windowLimits = self.get_gt_anomalies_segmentwise()
        dataSetName = "dummyname"
        anomalyList = self.sweeper.calcSweepScore(timestamps, anomaly_scores, windowLimits, dataSetName)
        scoresByThreshold = self.sweeper.calcScoreByThreshold(anomalyList)

        assert scoresByThreshold[0].threshold == 1.1  # all points regarded normal
        assert scoresByThreshold[1].threshold == 1.0  # anomal points regarded anomal

        return scoresByThreshold[0].score, scoresByThreshold[1].score


class Best_threshold_pw(Nonbinary_detection):
    def __init__(self, *args):
        self.name = "\\bestpwf"
        super().__init__(*args)

    def get_score(self):
        scores = []
        for current_anomaly_score in self.get_anomaly_score():
            scores.append(self.get_score_given_anomaly_score_and_threshold(threshold=current_anomaly_score))
        return np.nanmax(scores)

    def get_score_given_anomaly_score_and_threshold(self, threshold):
        gt = self.get_gt_anomalies_full_series()
        pred = np.array(self.get_anomaly_score()) >= threshold
        return f1_score(tp=pred @ gt, fn=(1 - pred) @ gt, fp=(1 - gt) @ pred)


class AUC_ROC(Nonbinary_detection):
    def __init__(self, *args):
        self.name = "\\aucroc"
        super().__init__(*args)

    def get_score(self):
        gt = self.get_gt_anomalies_full_series()
        return roc_auc_score(gt, self.get_anomaly_score())


class AUC_PR_pw(Nonbinary_detection):
    def __init__(self, *args):
        self.name = "\\aucpr"
        super().__init__(*args)

    def get_score(self):
        gt = self.get_gt_anomalies_full_series()
        return average_precision_score(gt, self.get_anomaly_score())


class VUS_ROC(Nonbinary_detection):
    def __init__(self, *args, max_window=4):
        super().__init__(*args)
        self.name = f"\\vusroc{{{max_window}}}"
        self.max_window = max_window

    def get_score(self):
        gt = np.array(self.get_gt_anomalies_full_series())
        score = np.array(self.get_anomaly_score())
        _, _, _, _, _, _, roc, pr = generate_curve(gt, score, self.max_window)
        return roc


class VUS_PR(Nonbinary_detection):
    def __init__(self, *args, max_window=4):
        super().__init__(*args)
        self.name = f"\\vuspr{{{max_window}}}"
        self.max_window = max_window

    def get_score(self):
        gt = np.array(self.get_gt_anomalies_full_series())
        score = np.array(self.get_anomaly_score())
        _, _, _, _, _, _, roc, pr = generate_curve(gt, score, self.max_window)
        return pr


class PatK_pw(Nonbinary_detection):
    def __init__(self, *args):
        super().__init__(*args)
        self.name = f"\\patk[{len(self.get_gt_anomalies_ptwise())}]"

    def get_score(self):
        gt = self.get_gt_anomalies_full_series()

        k = int(sum(gt))
        assert k > 0
        threshold = np.sort(self.get_anomaly_score())[-k]

        pred = self.get_anomaly_score() >= threshold
        assert sum(pred) >= k, (k, pred)

        return pred @ gt / sum(pred)



import numpy as np
import cv2 as cv
from copy import deepcopy
import pathlib

def convert_index(org_index, max_index, graph_width, margin_left):
    return round(float(org_index/max_index)*graph_width+margin_left)

def draw_csv(ranges, img, h_floor, h_ceiling, color, max_index, graph_width, margin_left):
    for a_range in ranges:
        start_time = convert_index(a_range.get_time()[0], max_index, graph_width, margin_left)
        end_time = convert_index(a_range.get_time()[1], max_index, graph_width, margin_left)
        cv.rectangle(img, (start_time, h_floor), (end_time, h_ceiling), color, thickness=-1)

def draw_csv_range(ranges, img, h_floor, h_ceiling, color, start, end):
    for a_range in ranges:
        if a_range.get_time()[0] <= end or a_range.get_time()[1] >= start:
            cv.rectangle(img, (a_range.get_time()[0]-start+10, h_floor), (a_range.get_time()[1]-start+10, h_ceiling), color, thickness=-1)

def shift_ranges(ranges, first_idx):
    for a_range in ranges:
        a_range.set_time(a_range.get_time()[0] - first_idx, a_range.get_time()[1] - first_idx)

def draw_graphs(anomalies, predictions, how_show: str):
    method_list = [ 'Anomalies', 'Predictions' ]
    anomalies = deepcopy(anomalies)
    predictions = deepcopy(predictions)
    first_idx = min(anomalies[0].get_time()[0]-100, predictions[0].get_time()[0])
    last_idx = max(anomalies[-1].get_time()[1], predictions[-1].get_time()[1])
    marginal_idx = int(float(last_idx-first_idx)/100)
    first_idx -= marginal_idx
    shift_ranges(anomalies, first_idx)
    shift_ranges(predictions, first_idx)
    ranges_list = [ anomalies, predictions ]
    max_index = max(anomalies[-1].get_time()[1], predictions[-1].get_time()[1]) + marginal_idx

    color_list = [(70, 70, 70),  #black
                  (60, 76, 203),  #red
                  (193, 134, 46),  #blue
                  (133, 160, 22),  #green
                  (206, 143, 187),  #purple
                  (94, 73, 52),  # darkblue
                  (63, 208, 244)  #yellow
                  ]

    margin_left   = 10
    margin_right  = 150
    margin_top    = 20
    margin_bottom = 40

    graph_gap = 20
    graph_height = 40
    graph_width = 2000

    n_results = 2

    width = margin_left + graph_width + margin_right
    height = margin_top + margin_bottom + n_results * (graph_gap + graph_height)
    bpp = 3

    img = np.ones((height, width, bpp), np.uint8)*255

    img_h = img.shape[0]
    img_w = img.shape[1]
    img_bpp = img.shape[2]

    thickness = 1
    fontsize = 1
    cv.line(img, (int(margin_left/2), img_h-margin_bottom), (img_w-int(margin_left/2), img_h-margin_bottom), color_list[0], thickness) #x-axis
    pts = np.array([[img_w-int(margin_left/2), img_h-margin_bottom], [img_w-int(margin_left/2)-7, img_h-margin_bottom+5], [img_w-int(margin_left/2)-7, img_h-margin_bottom-5]], np.int32) #arrow_head
    pts = pts.reshape((-1, 1, 2))
    cv.fillPoly(img, [pts], color_list[0])
    cv.putText(img, 'Relative Index', (img_w-180, img_h-15), cv.FONT_HERSHEY_COMPLEX_SMALL, fontsize, color_list[0], 1, cv.LINE_AA) #x-axis label

    for i in range(margin_left, width-margin_right, int(graph_width/10)):
        cv.line(img, (i, img_h-margin_bottom+2), (i, img_h-margin_bottom-2), color_list[0], thickness)
        org_index = str(round((i-10) / graph_width * max_index / 1000))
        cv.putText(img, org_index+'K', (i-len(org_index)*5, img_h-margin_bottom + 25), cv.FONT_HERSHEY_COMPLEX_SMALL, fontsize, color_list[0], 1, cv.LINE_AA)

    thickness = -1
    for idx in range(n_results):
        cv.putText(img, method_list[idx],
                   (width - margin_right + 2, img_h - margin_bottom - graph_gap * (idx+1) - graph_height * idx - 12),
                   cv.FONT_HERSHEY_COMPLEX_SMALL, fontsize, color_list[0], 1, cv.LINE_AA)
        draw_csv(ranges_list[idx], img, h_floor=img_h - margin_bottom - graph_gap * (idx+1) - graph_height * idx,
                 h_ceiling=img_h - margin_bottom - graph_gap * (idx+1) - graph_height * (idx+1),
                 color=color_list[(idx+1)%len(color_list)],
                 max_index=max_index, graph_width=graph_width, margin_left=margin_left)

    if how_show == 'screen' or how_show == 'all':
        cv.imshow("drawing", img)
    if how_show == 'file' or how_show == 'all':
        cv.imwrite("../../brief_result.png", img)
    if how_show != 'screen' and how_show != 'all' and how_show != 'file':
        print('Parameter Error')
    cv.waitKey(0);


def draw_multi_graphs(anomalies, predictions_list, predictions_name_list, how_show: str):
    method_list = [ 'Anomalies' ] + predictions_name_list

    anomalies = deepcopy(anomalies)
    predictions_list = deepcopy(predictions_list)

    first_idx = anomalies[0].get_time()[0]-100
    last_idx = anomalies[-1].get_time()[1]
    for single_prediction in predictions_list:
        first_idx = min(first_idx, single_prediction[0].get_time()[0])
        last_idx = max(last_idx, single_prediction[-1].get_time()[1])

    marginal_idx = int(float(last_idx-first_idx)/100)
    first_idx -= marginal_idx
    shift_ranges(anomalies, first_idx)
    for single_prediction in predictions_list:
        shift_ranges(single_prediction, first_idx)

    ranges_list = [ anomalies ] + predictions_list

    max_index = anomalies[-1].get_time()[1]
    for single_prediction in predictions_list:
        max_index = max(max_index, single_prediction[-1].get_time()[1])
    max_index = max_index + marginal_idx

    color_list = [(0, 0, 0),  #black
                  (60, 76, 203),  #red
                  (193, 134, 46),  #blue
                  (133, 160, 22),  #green
                  (206, 143, 187),  #purple
                  (94, 73, 52),  # darkblue
                  (63, 208, 244)  #yellow
                  ]

    margin_left   = 10
    margin_right  = 180
    margin_top    = 20
    margin_bottom = 40

    graph_gap = 20
    graph_height = 40
    graph_width = 2000

    n_results = len(ranges_list)

    width = margin_left + graph_width + margin_right
    height = margin_top + margin_bottom + n_results * (graph_gap + graph_height)
    bpp = 3

    img = np.ones((height, width, bpp), np.uint8)*255

    img_h = img.shape[0]
    img_w = img.shape[1]
    img_bpp = img.shape[2]

    thickness = 1
    fontsize = 1.4
    cv.line(img, (int(margin_left/2), img_h-margin_bottom), (img_w-int(margin_left/2), img_h-margin_bottom), color_list[0], thickness) #x-axis
    pts = np.array([[img_w-int(margin_left/2), img_h-margin_bottom], [img_w-int(margin_left/2)-7, img_h-margin_bottom+5], [img_w-int(margin_left/2)-7, img_h-margin_bottom-5]], np.int32) #arrow_head
    pts = pts.reshape((-1, 1, 2))
    cv.fillPoly(img, [pts], color_list[0])
    cv.putText(img, 'Relative Index', (img_w-180, img_h-15), cv.FONT_HERSHEY_COMPLEX_SMALL, 1, color_list[0], 1, cv.LINE_AA) #x-axis label

    for i in range(margin_left, width-margin_right, int(graph_width/10)):
        cv.line(img, (i, img_h-margin_bottom+2), (i, img_h-margin_bottom-2), color_list[0], thickness)
        org_index = str(round((i-10) / graph_width * max_index / 1000))
        cv.putText(img, org_index+'K', (i-len(org_index)*5, img_h-margin_bottom + 25), cv.FONT_HERSHEY_COMPLEX_SMALL, 1, color_list[0], 1, cv.LINE_AA)

    thickness = -1
    for idx in range(n_results):
        cv.putText(img, method_list[idx],
                   (width - margin_right + 2, img_h - margin_bottom - graph_gap * (idx+1) - graph_height * idx - 12),
                   cv.FONT_HERSHEY_COMPLEX_SMALL, fontsize, color_list[0], 1, cv.LINE_AA)
        draw_csv(ranges_list[idx], img, h_floor=img_h - margin_bottom - graph_gap * (idx+1) - graph_height * idx,
                 h_ceiling=img_h - margin_bottom - graph_gap * (idx+1) - graph_height * (idx+1),
                 color=color_list[(idx+1)%len(color_list)],
                 max_index=max_index, graph_width=graph_width, margin_left=margin_left)

    if how_show == 'screen' or how_show == 'all':
        cv.imshow("drawing", img)
    if how_show == 'file' or how_show == 'all':
        print("The file is saved at " + str(pathlib.Path(__file__).parent.absolute()))
        cv.imwrite("./brief_result.png", img)
    if how_show != 'screen' and how_show != 'all' and how_show != 'file':
        print('Parameter Error')
    cv.waitKey(0);



from typing import Union

import numpy as np

from ..base.time_series_metrics import BaseTimeSeriesMetrics


class TimeSeriesRecall(BaseTimeSeriesMetrics):
    """ This class calculates recall for time series """

    def __init__(self, alpha=0.0, cardinality="one", bias="flat"):
        """Constructor

        Args:
            alpha (float, optional): 0 <= alpha <= 1. Defaults to 0.0.
            cardinality (str, optional): ["one", "reciprocal", "udf_gamma"]. Defaults to "one".
            bias (str, optional): ["flat", "front", "middle", "back"]. Defaults to "flat".
        """

        assert (alpha >= 0) & (alpha <= 1)
        assert cardinality in ["one", "reciprocal", "udf_gamma"]
        assert bias in ["flat", "front", "middle", "back"]

        self.alpha = alpha
        self.cardinality = cardinality
        self.bias = bias

    def score(self, real: Union[np.ndarray, list], pred: Union[np.ndarray, list]) -> float:
        """Computing recall score

        Args:
            real (np.ndarray or list):
                One-dimensional array of correct answers with values of 1 or 0.
            pred (np.ndarray or list):
                One-dimensional array of predicted answers with values of 1 or 0.
        Returns:
            float: recall
        """

        assert isinstance(real, np.ndarray) or isinstance(real, list)
        assert isinstance(pred, np.ndarray) or isinstance(pred, list)

        if not isinstance(real, np.ndarray):
            real = np.array(real)
        if not isinstance(pred, np.ndarray):
            pred = np.array(pred)

        real_anomalies, predicted_anomalies = self._prepare_data(real, pred)
        recall = self._update_recall(real_anomalies, predicted_anomalies)

        return recall

    def _update_recall(self, real_anomalies: np.ndarray, predicted_anomalies: np.ndarray) -> float:
        """Update recall
        Args:
            real_anomalies (np.ndarray):
                2-dimensional array of the first and last indexes of each real anomaly range.
                e.g. np.array([[1933,  1953],[1958,  2000], ...])
            predicted_anomalies (np.ndarray):
                2-dimensional array of the first and last indexes of each predicted anomaly range.
                e.g. np.array([[1933,  1953],[1958,  2000], ...])
        Returns:
            float: recall
        """

        recall = 0
        if len(real_anomalies) == 0:
            return 0
        for i in range(len(real_anomalies)):
            omega_reward = 0
            overlap_count = [0]
            range_r = real_anomalies[i, :]
            for j in range(len(predicted_anomalies)):
                range_p = predicted_anomalies[j, :]
                omega_reward += self._compute_omega_reward(range_r, range_p, overlap_count)
            overlap_reward = self._gamma_function(overlap_count) * omega_reward

            if overlap_count[0] > 0:
                existence_reward = 1
            else:
                existence_reward = 0

            recall += self.alpha * existence_reward + (1 - self.alpha) * overlap_reward
        recall /= len(real_anomalies)
        return recall



# To store a single anomaly
class Range:
    def __init__(self, first, last, name):
        self._first_timestamp = first
        self._last_timestamp = last
        self._name = name

    def set_time(self, first, last):
        self._first_timestamp = first
        self._last_timestamp = last

    def get_time(self):
        return self._first_timestamp, self._last_timestamp

    def set_name(self, str):
        self._name = str

    def get_name(self):
        return self._name

    def get_len(self):
        return self._last_timestamp - self._first_timestamp + 1

    def __eq__(self, other):
        return self._first_timestamp == other.get_time()[0] and self._last_timestamp == other.get_time()[1]

    def distance(self, other_range) -> int:
        if min(self._last_timestamp, other_range.get_time()[1]) - max(self._first_timestamp, other_range.get_time()[0]) > 0:
            return 0
        else:
            return min(abs(self._first_timestamp - other_range.get_time()[1]),
                       abs(self._last_timestamp - other_range.get_time()[0]))

    def compare(self, other_range) -> int:
        if min(self._last_timestamp, other_range.get_time()[1]) - max(self._first_timestamp, other_range.get_time()[0]) > 0:
            return 0
        elif self._last_timestamp - other_range.get_time()[0] < 0:
            return -1
        else:
            return 1

def stream_2_ranges(self, prediction_stream: list) -> list:
    result = []
    for i in range(len(prediction_stream)-1):
        start_time = 0
        if prediction_stream[i] == 0 and prediction_stream[i+1] == 1:
            start_time = i+1
        elif prediction_stream[i] == 1 and prediction_stream[i+1] == 0:
            result.append(Range(start_time, i, ''))
    return result



# import numpy as np
# from matplotlib import pyplot as plt


# x = np.arange(48)

# y = np.sin(0.7 + x / 12) + np.sin(x / 4 + 29) + 0.1 * np.sin(1.25 * x) * (np.cos(np.sqrt(1.25 * x) + 2)) + x / 32 + 0.12


# figsize = (3.4, 2)
# plt.figure(figsize=figsize)

# plt.plot(x, y)

# for t in [0.5, 1, 1.5, 2, 2.5]:
#     plt.plot(x, x + t - x, ".", color="dimgray")
#     for i in range(len(x)):
#         if t < y[i]:
#             plt.plot([x[i]], [t], ".r")
#             plt.plot([x[i]], [t], "xw", markersize=2)

# fs = 7
# plt.xlabel("Time", fontsize=fs)
# plt.ylabel("Anomaly score / Threshold", fontsize=fs)
# plt.xticks(fontsize=fs)
# plt.yticks(fontsize=fs)

# plt.tight_layout()
# plt.savefig("thr2.pdf")
# plt.show()



# import numpy as np
# from scipy.stats import norm
# from matplotlib import pyplot as plt

# from metrics import f1_from_pr


# class Two_1d_normal_distributions:
#     def __init__(self, P_ampl, N_ampl, P_mu, N_mu, P_std, N_std, color="b", betas=None):
#         self.P_ampl = P_ampl
#         self.N_ampl = N_ampl
#         self.P_mu = P_mu
#         self.N_mu = N_mu
#         self.P_std = P_std
#         self.N_std = N_std
#         self.betas = (1 / 8, 1 / 4, 1 / 2, 1, 2, 4, 8, 16, 32) if betas == None else betas

#         self.color = color
#         self.N_color = "k"
#         self.P_color = "r"

#     def make(self, delta=0.05, steps=10001, start=-8, stop=8):
#         index = 0

#         # For plotting grahps
#         self.fpr = []
#         self.precision = []
#         self.recall = []

#         # For plotting x´s and o´s on the graphs
#         self.x_fpr = []
#         self.x_precision = []
#         self.x_recall = []
#         self.xs = 0
#         self.x_threshold = []
#         self.o_fpr = []
#         self.o_precision = []
#         self.o_recall = []
#         self.os = 0
#         self.o_threshold = []

#         # Track maximum f scores for various beta values
#         self.max_f = {i: 0 for i in self.betas}
#         self.max_f_fpr = {i: 0 for i in self.betas}
#         self.max_f_precision = {i: 0 for i in self.betas}
#         self.max_f_recall = {i: 0 for i in self.betas}
#         self.max_f_thresholds = {i: 0 for i in self.betas}

#         for threshold in np.linspace(start, stop, steps):
#             TN = self.N_ampl * norm.cdf(threshold, loc=self.N_mu, scale=self.N_std)
#             FP = self.N_ampl - TN
#             FN = self.P_ampl * norm.cdf(threshold, loc=self.P_mu, scale=self.P_std)
#             TP = self.P_ampl - FN
#             self.fpr.append(FP / (FP + TN))
#             self.precision.append(TP / (TP + FP))
#             self.recall.append(TP / (TP + FN))

#             for beta in self.betas:
#                 if f1_from_pr(p=self.precision[-1], r=self.recall[-1], beta=beta) > self.max_f[beta]:
#                     self.max_f[beta] = f1_from_pr(p=self.precision[-1], r=self.recall[-1], beta=beta)
#                     self.max_f_fpr[beta] = self.fpr[-1]
#                     self.max_f_precision[beta] = self.precision[-1]
#                     self.max_f_recall[beta] = self.recall[-1]
#                     self.max_f_thresholds[beta] = threshold

#             if (FN) / (self.P_ampl) >= self.xs * delta + delta * 0.5:
#                 self.x_fpr.append(FP / (FP + TN))
#                 self.x_precision.append(TP / (TP + FP))
#                 self.x_recall.append(TP / (TP + FN))
#                 self.xs += 1
#                 # print(xs, TN+FN)
#                 self.x_threshold.append(threshold)
#             if (TN / self.N_ampl) >= self.os * delta + delta * 0.5:
#                 self.o_fpr.append(FP / (FP + TN))
#                 self.o_precision.append(TP / (TP + FP))
#                 self.o_recall.append(TP / (TP + FN))
#                 self.os += 1
#                 self.o_threshold.append(threshold)
#                 # print(os, TN+FN)

#     def plot_roc_pr(self, roc_ax, pr_ax, plot_xs=True, plot_os=True, plot_fs=False):
#         roc_ax.plot(self.fpr, self.recall, self.color, zorder=1)
#         pr_ax.plot(self.precision, self.recall, self.color, zorder=1)
#         if plot_xs:
#             roc_ax.plot(self.x_fpr, self.x_recall, "x", color=self.color, zorder=1)
#             pr_ax.plot(self.x_precision, self.x_recall, "x", color=self.color, zorder=1)
#         if plot_os:
#             roc_ax.plot(self.o_fpr, self.o_recall, "o", color=self.color, fillstyle="none", zorder=1)
#             pr_ax.plot(self.o_precision, self.o_recall, "o", color=self.color, fillstyle="none", zorder=1)
#         if plot_fs:
#             roc_ax.plot(
#                 list(self.max_f_fpr.values()),
#                 list(self.max_f_recall.values()),
#                 ".",
#                 linestyle="None",
#                 zorder=2,
#                 color="k",
#             )  # self.color)
#             pr_ax.plot(
#                 list(self.max_f_precision.values()), list(self.max_f_recall.values()), ".", zorder=2, color="k"
#             )  # self.color)
#             for beta in self.betas:
#                 # roc_ax.plot([self.max_f_fpr[beta]],[self.max_f_recall[beta]], marker=f"$1/{int(1/beta)}$" if beta<1 else f"${beta}$", linestyle= "None", zorder=2, color="k")
#                 # pr_ax.plot([self.max_f_precision[beta]], [self.max_f_recall[beta]], marker=f"$1/{int(1/beta)}$" if beta<1 else f"${beta}$", zorder=2, color="k")
#                 if self.color == "forestgreen":  # need to place the numbers differently
#                     roc_ax.text(
#                         self.max_f_fpr[beta],
#                         self.max_f_recall[beta],
#                         f" $1/{int(1/beta)}$" if beta < 1 else f" ${beta}$",
#                         horizontalalignment="left",
#                         verticalalignment="top",
#                         color=self.color,
#                     )
#                     pr_ax.text(
#                         self.max_f_precision[beta],
#                         self.max_f_recall[beta],
#                         f"$1/{int(1/beta)}$" if beta < 1 else f"${beta}$",
#                         horizontalalignment="left",
#                         verticalalignment="bottom",
#                         color=self.color,
#                     )
#                 else:
#                     roc_ax.text(
#                         self.max_f_fpr[beta],
#                         self.max_f_recall[beta],
#                         f"$1/{int(1/beta)}$" if beta < 1 else f"${beta}$",
#                         horizontalalignment="right",
#                         verticalalignment="bottom",
#                         color=self.color,
#                     )
#                     pr_ax.text(
#                         self.max_f_precision[beta],
#                         self.max_f_recall[beta],
#                         f"$1/{int(1/beta)}$ " if beta < 1 else f"${beta}$ ",
#                         horizontalalignment="right",
#                         verticalalignment="top",
#                         color=self.color,
#                     )

#             # adjust axes to get the numbers within the figure
#             xmin, xmax = roc_ax.get_xlim()
#             xmin, xmax = pr_ax.get_xlim()
#             roc_ax.set_xlim([xmin - 0.04, xmax])
#             pr_ax.set_xlim([xmin - 0.01, xmax])
#         pr_ax.set_xlabel("Precision")
#         pr_ax.set_ylabel("Recall")
#         roc_ax.set_xlabel("False positive rate")
#         roc_ax.set_ylabel("Recall")

#     def plot_roc_prec(self):
#         plt.plot(self.fpr, self.recall, self.color, zorder=1)
#         plt.plot(self.fpr, self.precision, self.color, zorder=1)
#         plt.show()

#     def plot_roc_pr_lines(self, ax):
#         ax.plot(self.fpr, np.array(self.recall) + 1, self.color, zorder=1)
#         ax.plot(np.array(self.precision) * (-1) + 1, self.recall, self.color, zorder=1)
#         for i in range(0, len(self.recall), 5):
#             ax.plot(
#                 [self.fpr[i], 1 - self.precision[i]],
#                 [self.recall[i] + 1, self.recall[i]],
#                 marker="o",
#                 color=self.N_color,
#                 zorder=1,
#                 alpha=0.3,
#             )

#     def plot_distributions(
#         self, axes, start=-5, stop=7, steps=1001, normalize=True, plot_xs=True, plot_os=True, plot_fs=False, threshold=0
#     ):
#         grid = np.linspace(start, stop, steps)
#         fill_alpha = 0.2

#         y = lambda x: norm.pdf(x, loc=self.N_mu, scale=self.N_std) * (1 if normalize else self.N_ampl)
#         axes[0].plot(
#             grid,
#             y(grid),
#             color=self.N_color,
#             label=f"pdf_N/{self.N_ampl}",
#         )

#         axes[0].fill_between(
#             grid[grid <= threshold], 0, y(grid[grid <= threshold]), alpha=fill_alpha, lw=0, color="darkgreen"
#         )
#         axes[0].fill_between(
#             grid[grid >= threshold], 0, y(grid[grid >= threshold]), alpha=fill_alpha, lw=0, color="orchid"
#         )
#         tn_x = min(self.N_mu, threshold - 0.75)
#         fp_x = max(self.N_mu, threshold + 0.75)
#         axes[0].text(tn_x, y(tn_x) / 2 - 0.005, "TN", horizontalalignment="center", verticalalignment="top")
#         axes[0].text(fp_x, y(fp_x) / 2 - 0.005, "FP", horizontalalignment="center", verticalalignment="top")
#         # add thresholdline, on the whole y-range
#         ymin, ymax = axes[0].get_ylim()
#         axes[0].plot([threshold, threshold], [ymin - 1, ymax + 1], "--", color="gray", lw=1)
#         axes[0].set_ylim([ymin - 0.02, ymax])

#         # same for anomal distributions
#         y = lambda x: norm.pdf(x, loc=self.P_mu, scale=self.P_std) * (1 if normalize else self.P_ampl)
#         axes[1].plot(
#             grid,
#             y(grid),
#             color=self.P_color,
#             label=f"pdf_P/{self.P_ampl}",
#         )
#         axes[1].fill_between(
#             grid[grid <= threshold], 0, y(grid[grid <= threshold]), alpha=fill_alpha, lw=0, color="chocolate"
#         )
#         axes[1].fill_between(
#             grid[grid >= threshold], 0, y(grid[grid >= threshold]), alpha=fill_alpha, lw=0, color="darkcyan"
#         )
#         fn_x = min(self.P_mu, threshold - 0.75)
#         tp_x = max(self.P_mu, threshold + 0.75)
#         axes[1].text(fn_x, y(fn_x) / 2, "FN", horizontalalignment="center", verticalalignment="top")
#         axes[1].text(tp_x, y(tp_x) / 2, "TP", horizontalalignment="center", verticalalignment="top")

#         # add thresholdline, on the whole y-range
#         ymin, ymax = axes[1].get_ylim()
#         axes[1].plot([threshold, threshold], [ymin - 1, ymax + 1], "--", color="gray", lw=1)
#         axes[1].set_ylim([ymin, ymax])


#     def plot_cdf(self, ax, start=-6, stop=8, steps=1001, normalize=True):
#         grid = np.linspace(start, stop, steps)

#         ax.plot(
#             grid,
#             norm.cdf(grid, loc=self.N_mu, scale=self.N_std) * (1 if normalize else self.N_ampl),
#             color=self.N_color,
#             label=f"pdf_N/{self.N_ampl}",
#         )
#         ax.plot(
#             grid,
#             norm.cdf(grid, loc=self.P_mu, scale=self.P_std) * (1 if normalize else self.P_ampl),
#             color=self.P_color,
#             label=f"pdf_P/{self.P_ampl}",
#         )


# if __name__ == "__main__":

#     # Make detector distributions

#     t1 = Two_1d_normal_distributions(
#         1, 49, 1.8, -1, 2, 1, color="mediumblue", betas=(1 / 8, 1 / 4, 1 / 2, 1, 2, 4, 8, 16)
#     )
#     t2 = Two_1d_normal_distributions(
#         1, 49, 1, -1, 1, 1, color="forestgreen", betas=(1 / 8, 1 / 4, 1 / 2, 1, 2, 4, 8, 16)
#     )

#     t1.make(steps=1001, delta=0.1)
#     t2.make(steps=1001, delta=0.1)

    
#     #  Make roc and pr plots

#     figsize = (4, 4)

#     roc_fig, roc_ax = plt.subplots(figsize=figsize)
#     pr_fig, pr_ax = plt.subplots(figsize=figsize)
#     t1.plot_roc_pr(roc_ax, pr_ax, False, False, True)
#     t2.plot_roc_pr(roc_ax, pr_ax, False, False, True)
#     roc_fig.tight_layout()
#     pr_fig.tight_layout()
#     roc_fig.savefig("auc_roc_f.pdf")
#     pr_fig.savefig("auc_pr_f.pdf")
#     plt.show()
#     plt.close("all")


#     # Make distribution plots

#     figsize = (5, 3)

#     for beta in t1.betas:  # [16,4,1,1/4,1/8]:
#         fig, axes = plt.subplots(2, 2, figsize=figsize, sharex=True, sharey=True)
#         t1.plot_distributions(
#             [axes[0][0], axes[1][0]], plot_xs=False, plot_os=False, plot_fs=True, threshold=t1.max_f_thresholds[beta]
#         )
#         t2.plot_distributions(
#             [axes[0][1], axes[1][1]], plot_xs=False, plot_os=False, plot_fs=True, threshold=t2.max_f_thresholds[beta]
#         )

#         axes[0][0].set_title(f"Blue detector", color=t1.color)
#         axes[0][1].set_title("Green detector", color=t2.color)
#         axes[1][0].set_xlabel("Anomaly \n score", color=t1.color)
#         axes[1][1].set_xlabel("Anomaly \n score", color=t2.color)
#         shadowaxes = fig.add_subplot(111, xticks=[], yticks=[], frame_on=False)
#         shadowaxes.set_ylabel("Probability density", labelpad=25)
#         fig.tight_layout()
#         axes[0][0].set_ylabel("Normal\nsamples", labelpad=25)
#         axes[1][0].set_ylabel("Anomalous\nsamples", labelpad=25)

#         plt.subplots_adjust(hspace=0.0)
#         plt.savefig(f"auc_distributions_b{beta}.pdf")
#         plt.show()
